library state_groups;

import 'package:flutter/widgets.dart';

/// Make your state inherit from [SyncState] instead of [State]. That will allow
/// this class to do all the message handling stuff for you. Note that update
/// and shouldUpdate() can be overridden for custom behaviour.
abstract class SyncState<MessageType, T extends StatefulWidget> extends State<T> {

	// This causes a runtime error if a stategroup with the wrong generic type
	// is given, not a compile time error ¯\_(ツ)_/¯
	SyncState(StateGroup<MessageType> group, {this.errorFunction}) {
		_groups.add(group);
	}

	/// This is only for legacy reasons. Ignore it.
	SyncState.Advanced(this._groups, {this.errorFunction});

	List<StateGroup<MessageType>> _groups = <StateGroup<MessageType>>[];
	Function(String)? errorFunction;

	@override
	void initState() {
		for (int i = 0; i < _groups.length; i++) {
			_groups[i].states.add(this);
		}
		super.initState();
	}

	@override
	void dispose() {
		for (int i = 0; i < _groups.length; i++) {
			if (!_groups[i].states.remove(this)) {
				errorFunction!("Could not find $this in state group ${_groups[i]}]");
			}
		}
		super.dispose();
	}

	/// By default this calls setState() if we should update. Override it for
	/// custom behaviour.
	void update(MessageType? message) {
		if (!shouldUpdate(message)) {
			return;
		}
		WidgetsBinding.instance!.addPostFrameCallback((_) {
			if (mounted) {
				setState(() {});
			}
		});
		WidgetsBinding.instance!.scheduleFrame();
	}

	/// This is checked by update() and not notifyAll() in the [StateGroup] I may
	/// change it in the future
	bool shouldUpdate(MessageType? message) {
		return true;
	}
}

/// A collection of [SyncState] objects. Use notifyAll() to notify them all.
class StateGroup<MessageType> {

	StateGroup({this.onNotify});

	final void Function(MessageType?)? onNotify;
	final Set<SyncState<MessageType, dynamic>> states = <SyncState<MessageType, dynamic>>{};

	/// Updates all states with the given message
	void notifyAll([MessageType? message]) {
		if (onNotify != null) {
			onNotify!(message);
		}
		for (final SyncState<MessageType, dynamic> state in states) {
			state.update(message);
		}
	}

	@override
	String toString() {
		return "State Group ($states)";
	}
}

/// A collection of [StateGroup] objects. Use notifyAll() to notify them all.
class StateGroupCollection<MessageType> {

	StateGroupCollection(this.collections);

	final List<StateGroup<MessageType>> collections;

	/// Updates all states with the given message
	void notifyAll([MessageType? message]) {
		for (int i = 0; i < collections.length; i++) {
			collections[i].notifyAll(message);
		}
	}

	@override
	String toString() {
		return "State Group Collection: ($collections)";
	}
}
